from .types import Annotations, Meta, Input, Output, CorrectResult, BadlyPreprocessed, Item, \
  Message, Done, Preds
from .sdk import SDK
from ._api import api
from .pipeline import Context, correct

__all__ = [
  'Annotations', 'Meta', 'Input', 'Output', 'CorrectResult', 'BadlyPreprocessed', 'Item',
  'Message', 'Done', 'Preds',
  'SDK', 'api',
  'Context', 'correct',
]


from typing import Sequence
from dataclasses import dataclass
from kv import KV
from pipeteer.queues import ReadQueue, WriteQueue, Transaction, InexistentItem
import sequence_edits as se
import game_prediction as gp
from moveread.pipelines.correct import Annotations, Meta, Input, Output, CorrectResult, BadlyPreprocessed, Item

@dataclass
class SDK:
  Qin: ReadQueue[Input]
  Qout: WriteQueue[Output]
  cache: KV[Annotations]

  async def items(self):
    async for e in self.Qin.items():
      yield Meta.of(e)

  async def item(self, id: str):
    inp = await self.Qin.safe_read(id)
    if inp:
      anns = await self.cache.safe_read(id)or Annotations()
      return Item(title=inp['title'], boxes=inp['boxes'], anns=anns)
  
  async def confirm(self, id: str, pgn: Sequence[str], early: bool):
    try:
      anns = await self.cache.safe_read(id) or Annotations()
      result = CorrectResult(pgn=pgn, early=early, annotations=anns)
      async with Transaction(self.Qin, self.Qout, autocommit=True):
        await self.Qout.push(id, result)
        await self.Qin.pop(id)
      return True
    except InexistentItem:
      return False

  async def repreprocess(self, id: str):
    try:
      async with Transaction(self.Qin, self.Qout, autocommit=True):
        await self.Qin.pop(id)
        await self.Qout.push(id, BadlyPreprocessed())
      return True
    except InexistentItem:
      return False

  async def annotate(self, id: str, anns: Annotations):
    return await self.cache.insert(id, anns)
  
  async def predict(self, id: str, *, fen: str, beam_width: int = 4):
    task = await self.Qin.read(id)
    anns = await self.cache.safe_read(id) or Annotations()
    preds = task['ocrpreds'][:(anns and anns.end_correct)]
    preds = se.apply(anns.edits or [], list(preds), fill=[('', 0)])

    return gp.guided_predict(
      preds, manual_ucis=anns.manual_ucis, fen=fen, beam_width=beam_width,
      motions=anns.motions(), languages=anns.langs()
    )


from typing_extensions import Literal, Sequence, TypedDict
from dataclasses import dataclass, field
from pydantic import BaseModel
import sequence_edits as se
from chess_notation import Language, PawnCapture, PieceCapture, MotionStyles

@dataclass
class ManualPred:
  san: str
  tag: Literal['manual'] = 'manual'

@dataclass
class AutoPred:
  san: str
  prob: float
  tag: Literal['predicted'] = 'predicted'
  
Pred = ManualPred | AutoPred

class Preds(BaseModel):
  reqId: str
  tag: Literal['preds'] = 'preds'
  preds: Sequence[Pred]

class Done(BaseModel):
  reqId: str
  tag: Literal['done'] = 'done'

Message = Preds | Done

NA = Literal['N/A']
def no_na(value):
  if value != 'N/A':
    return value

@dataclass
class Annotations:
  lang: Language | NA | None = None
  pawn_capture: PawnCapture | NA | None = None
  piece_capture: PieceCapture | NA | None = None
  end_correct: int | None = None
  manual_ucis: dict[int, str] = field(default_factory=dict)
  edits: Sequence[se.Edit] = field(default_factory=list)

  def motions(self) -> MotionStyles:
    styles = MotionStyles()
    if (pawn_cap := no_na(self.pawn_capture)) is not None:
      styles.pawn_captures = [pawn_cap]
    if (piece_cap := no_na(self.piece_capture)) is not None:
      styles.piece_captures = [piece_cap]
    return styles
  
  def langs(self) -> list[Language]:
    if (l := no_na(self.lang)) is not None:
      return [l]
    else:
      return ['EN', 'CA']

@dataclass
class Meta:
  title: str
  id: str
  @classmethod
  def of(cls, entry: tuple[str, 'Input']):
    k, v = entry
    return cls(id=k, title=v['title'])

class Input(TypedDict):
  title: str
  boxes: Sequence[str]
  ocrpreds: Sequence[Sequence[tuple[str, float]]]
  """MOVE x TOP_PREDS x (word, logprob)"""

@dataclass
class Item:
  title: str
  boxes: Sequence[str]
  anns: Annotations

@dataclass
class CorrectResult:
  annotations: Annotations
  pgn: Sequence[str]
  early: bool
  tag: Literal['correct'] = field(default='correct', kw_only=True)

@dataclass
class BadlyPreprocessed:
  tag: Literal['badly-preprocessed'] = 'badly-preprocessed'

Output = CorrectResult | BadlyPreprocessed



from dataclasses import dataclass
from pipeteer import task, ReadQueue, WriteQueue, Context as BaseContext
from kv import KV, LocatableKV
from moveread.pipelines.correct import Input, Output, Annotations

@dataclass
class Context(BaseContext):
  cache: KV[Annotations]
  blobs: LocatableKV[bytes]

@task()
def correct(Qin: ReadQueue[Input], Qout: WriteQueue[Output], ctx: Context):
  from moveread.pipelines.correct import api, SDK
  sdk = SDK(Qin, Qout, ctx.cache)
  return api(sdk, ctx.blobs, logger=ctx.log)



from typing import Generic, TypeVar
from dataclasses import dataclass, field
from haskellian import ManagedAsync

T = TypeVar('T')

@dataclass
class Streams(Generic[T]):
  channels: dict[str, ManagedAsync[T]] = field(default_factory=dict)

  def listen(self, id: str) -> ManagedAsync[T]:
    if not id in self.channels:
      self.channels[id] = ManagedAsync()
    return self.channels[id]
  
  def push(self, id: str, value: T):
    self.listen(id).push(value)

  def delete(self, id: str):
    if id in self.channels:
      del self.channels[id]

  def clear(self):
    self.channels.clear()


import os
from argparse import ArgumentParser

def env(variable: str, *, default = None, required: bool = True) -> dict:
  if (value := os.getenv(variable, default)) is not None:
    return dict(default=value)
  return dict(required=required)

def main():
  parser = ArgumentParser()
  parser.add_argument('--db', type=str, **env('DB_URL'), help='DB connection string')
  parser.add_argument('--cache', type=str, **env('CACHE_URL'), help='Cache connection string')
  parser.add_argument('--blobs', type=str, **env('BLOBS_URL'), help='Blobs KV connection string')
  parser.add_argument('--url', type=str, **env('SELF_URL'), help='Self URL')
  parser.add_argument('--secret', type=str, **env('SECRET'), help='Authorization secret')

  parser.add_argument('--host', type=str, default='0.0.0.0')
  parser.add_argument('-p', '--port', type=int, default=8000)

  args = parser.parse_args()

  import uvicorn
  from fastapi import Request, Response
  from fastapi.middleware.cors import CORSMiddleware
  from kv import KV, LocatableKV, ServerKV
  from pipeteer import Backend
  from moveread.pipelines.correct import Context, correct, Annotations

  cache = KV.of(args.cache, Annotations)
  blobs = KV.of(args.blobs, bytes)
  if not isinstance(blobs, LocatableKV):
    blobs = blobs.served(args.url.rstrip('/') + '/blobs', secret=args.secret)

  backend = Backend.sql(sql_url=args.db, url=args.url, secret=args.secret)
  ctx = Context(backend, cache=cache, blobs=blobs)

  backend.mount(correct, ctx)

  api = correct.run(ctx)
  @api.middleware('http')
  async def check_token(req: Request, call_next):
    if (
      req.headers.get('Authorization') == f'Bearer {args.secret}'
      or req.url.path.startswith('/blobs/')
      or req.url.path.startswith('/api/preds/')
    ):
      return await call_next(req)
    else:
      return Response(status_code=401, content='Unauthorized')
    
  backend.app.add_middleware(CORSMiddleware, allow_origins=['*'], allow_credentials=True, allow_methods=['*'], allow_headers=['*'])
  backend.app.mount('/blobs', ServerKV(blobs, type=bytes, secret=args.secret))
  backend.app.mount('/api', api)

  uvicorn.run(backend.app, host=args.host, port=args.port)


if __name__ == '__main__':
  import sys
  os.chdir('/home/m4rs/mr-github/rnd/pipelines-v2')
  sys.argv.extend([
    '--db', 'sqlite+aiosqlite:///docker/data/correct.db',
    '--blobs', 'file://docker/data/blobs',
    '--url', 'http://localhost:8003',
    '-p', '8003',
    '--secret', 'secret',
  ])
  main()


import asyncio
from pydantic import BaseModel
from kv import LocatableKV
from fastapi import FastAPI, Response
from sse_starlette import EventSourceResponse
from dslog import Logger
from dslog.uvicorn import setup_loggers_lifespan, DEFAULT_FORMATTER, ACCESS_FORMATTER
from moveread.pipelines.correct import SDK, Message, Meta, Item, Annotations, Done, Preds
from moveread.pipelines.correct.util import Streams

def api(
  sdk: SDK, images: LocatableKV[bytes], *,
  logger = Logger.click().prefix('[GAME CORRECTION]')
):

  streams = Streams[Message]()
  reqIds: dict[str, str] = {}

  app = FastAPI(
    generate_unique_id_function=lambda route: route.name,
    lifespan=setup_loggers_lifespan(
      access=logger.format(ACCESS_FORMATTER),
      uvicorn=logger.format(DEFAULT_FORMATTER),
    )
  )
  
  @app.get('/items', response_model_exclude_none=True)
  async def get_items() -> list[Meta]:
    return [x async for x in sdk.items()]
  
  @app.get('/item', response_model_exclude_none=True)
  async def get_item(id: str, resp: Response) -> Item | None:
    if (item := (await sdk.item(id))):
      item.boxes = [images.url(box) for box in item.boxes]
      return item
    else:
      resp.status_code = 404

  class ConfirmParams(BaseModel):
    pgn: list[str]
    early: bool
    
  @app.post('/confirm')
  async def confirm(id: str, params: ConfirmParams, resp: Response) -> bool:
    ok = await sdk.confirm(id, pgn=params.pgn, early=params.early)
    if not ok:
      resp.status_code = 404
    return ok
  
  @app.post('/repreprocess')
  async def repreprocess(id: str, resp: Response) -> bool:
    ok = await sdk.repreprocess(id)
    if not ok:
      resp.status_code = 404
    return ok
  
  @app.post('/annotate')
  async def annotate(id: str, anns: Annotations, resp: Response) -> bool:
    await sdk.annotate(id, anns)
    return True

  @app.post('/predict')
  async def predict(id: str, userId: str, reqId: str, fen: str) -> list[Preds|Done]:
    reqIds[userId] = reqId
    preds = []

    for ps in await sdk.predict(id, fen=fen):
      if reqIds.get(userId) != reqId:
        logger(f'UserID: {userId}, ReqId: {reqId} got canceled.', level='DEBUG')
        return preds
      msg = Preds(reqId=reqId, preds=ps) # type: ignore
      streams.push(userId, msg)
      await asyncio.sleep(0) # yield control to the event loop, to force result streaming
      preds.append(msg)
    
    streams.push(userId, Done(reqId=reqId))
    return preds

  @app.get('/preds/{userId}')
  def preds(userId: str):
    return EventSourceResponse(streams.listen(userId).map(lambda m: m.model_dump_json()))
  
  return app


# def clientgen():
#   from argparse import ArgumentParser

#   parser = ArgumentParser()
#   parser.add_argument('-o', '--output', required=True)

#   args = parser.parse_args()

#   from openapi_ts import generate_client

#   schema = fastapi({}, {}).openapi() # type: ignore
#   generate_client(schema, args.output)